package lazabs.ai

import lazabs.lattice.{Lattice, PointwiseLattice}
import lazabs.viewer.ScalaPrinter
import lazabs.cfg._
import lazabs.ast.ASTree._

object Signs {
 
  // A 5-element abstraction of sets of integers
  sealed abstract class Sign
  case object TopSign extends Sign    // all integers
  case object BottomSign extends Sign // {}
  case object Positive extends Sign   // {1,2,3,...}
  case object Zero extends Sign       // {0}
  case object Negative extends Sign   // {...,-3,-2,-1}
 
  class SignLattice extends Lattice { // Sign as a lattice
    type Elem = Sign
    val top = TopSign        // universal set is top
    val bottom = BottomSign  // empty set is bottom
    def leq(x : Sign, y : Sign) : Boolean = (x,y) match {
      case (BottomSign,_) => true  // empty set subset of all
      case (_,TopSign) => true     // full set superset of all
      case _ if x==y => true       // lattice order is reflexive
      case _ => false              // others turn out to be incomparable 
    }
    def join(x : Sign, y : Sign) : Sign = (x,y) match {
      case (BottomSign,_) => y  // bottom is neutral
      case (_,BottomSign) => x  //   and join is commutative
      case _ if x==y => x       // join is idempotent
      case _ => TopSign         // this lattice is very shallow; we get top easily
    }
    def meet(x : Sign, y : Sign) : Sign = (x,y) match {
      case (TopSign,_) => y  // dual to join
      case (_,TopSign) => x
      case _ if x==y => x
      case _ => BottomSign
    }
  }
 
  // Compute effect of statements on lattice elements.
  //   This is the key part of abstract interpreter
  object SignTransFun extends TransferFunction[Map[String,Sign]] {
    def apply(l : CFGLabel, fact : Map[String,Sign]) : Map[String,Sign] = {
      def evalExpr(e: Expression) : Sign = e match {
        case NumericalConst(n) => {
          if (n < 0) Negative
          else if (n > 0) Positive
          else Zero
        }
        case BoolConst(b) => {
          if (b) Positive
          else Negative
        }
        case Variable(v,_) => fact.getOrElse(v,BottomSign)
        case Subtraction(e1, e2) => (evalExpr(e1), evalExpr(e2)) match {
          case (_,BottomSign) => BottomSign
          case (BottomSign,_) => BottomSign
          case (TopSign,_) => TopSign
          case (_,TopSign) => TopSign
          case (Zero,y) => y
          case (x,Zero) => x
          case (Positive,Positive) => TopSign
          case (Negative,Negative) => TopSign
          case (Positive,Negative) => Positive
          case (Negative,Positive) => Negative
        }
        case Addition(e1, e2) => (evalExpr(e1), evalExpr(e2)) match {
          case (_,BottomSign) => BottomSign
          case (BottomSign,_) => BottomSign
          case (TopSign,_) => TopSign
          case (_,TopSign) => TopSign
          case (Zero,y) => y
          case (x,Zero) => x
          case (Positive,Positive) => Positive
          case (Negative,Negative) => Negative
          case (Positive,Negative) => TopSign
          case (Negative,Positive) => TopSign
        }
        case Equality(e1, e2) => (evalExpr(e1), evalExpr(e2)) match {
          case (_,BottomSign) => BottomSign
          case (BottomSign,_) => BottomSign
          case (Zero,Zero) => Positive /* concrete semantics gives 1 */
          case (Zero,Positive) => Zero /* cannot be equal, concrete gives 0 */
          case (Positive,Zero) => Zero 
          case (Negative,Zero) => Zero
          case (Zero,Negative) => Zero
          case (Positive,Negative) => Zero
          case (Negative,Positive) => Zero
          case (_,_) => TopSign /* who knows, e.g. two positive cound be same or not */
        }
        case Not(e) => evalExpr(e) match {
          case BottomSign => BottomSign
          case Positive => Zero
          case Zero => Positive
          case _ => sys.error("Not supported: " + ScalaPrinter(e))
        }
        case LessThanEqual(e1, e2) => (evalExpr(e1), evalExpr(e2)) match {
          case (_,BottomSign) => BottomSign
          case (BottomSign,_) => BottomSign
          case (Zero,Zero) => Positive /* concrete semantics gives 1 */
          case (Zero,Positive) => Positive
          case (Negative,Zero) => Positive
          case (Negative,Positive) => Positive
          case (Positive,Zero) => Zero /* cannot hold */
          case (Zero,Negative) => Zero
          case (Positive,Negative) => Zero
          case (_,_) => TopSign /* who knows */
        }
        case LessThan(e1, e2) => (evalExpr(e1), evalExpr(e2)) match {
          case (_,BottomSign) => BottomSign
          case (BottomSign,_) => BottomSign
          case (Zero,Zero) => Zero
          case (Zero,Positive) => Positive
          case (Negative,Zero) => Positive
          case (Negative,Positive) => Positive
          case (Positive,Zero) => Zero /* cannot hold */
          case (Zero,Negative) => Zero
          case (Positive,Negative) => Zero
          case (_,_) => TopSign /* who knows */
        }
        case _ => sys.error("Expression not supported: " + ScalaPrinter(e))
      }

      l match {
        case Assign(Variable(v,_), rhs) => fact.updated(v, evalExpr(rhs))
        case Assume(expr) => fact //sound but could be more precise. To complete.
        case Havoc(Variable(v,_)) => fact.updated(v, TopSign)
      }
    }
  }
   
  val signLattice = new SignLattice
  class PSignLattice(domain : Set[String]) extends 
    PointwiseLattice[String, Sign, SignLattice](signLattice, domain)
}
