package lazabs.cfg

import lazabs.digraph.Vertex
import lazabs.ast.ASTree._
import lazabs.ast.Annotation
import lazabs.utils.Manip._

object MakeCFG {

  val cfg = new CFG()
  val end = cfg.newVertex // the end state

  /**
   * gets the body of the given function and its set of arguments.
   */
  def getFuncExpression(fun: String, defs: List[Declaration]): (Expression,Set[String]) = {
    def getFuncBody(defs: List[Declaration]): (Expression,Set[String]) = defs match {
      case Nil =>
          throw new RuntimeException("Error: Could not find method " + fun)
      case (f@FunctionDefinition(name, params, _, body)) :: ds if (fun == name) => {
        f.getAnnotation match {
          case None =>
          case Some(Annotation("predicates", ps)) => cfg.predicates ++= ps.toSet
        }
        (body, params.map(_.name).toSet)
      }
      case _ :: ds => getFuncBody(ds)
    }
    getFuncBody(defs)
  }
  
  /**
   * Gets a Scala object and generates the control flow graph of the specified function
   */
  def apply(o: Sobject, fun: String): CFG = {
    val (funcExpression, params) = getFuncExpression(fun,o.defs)
    cfg.variables ++= params
    cfg.start = makeCFG(funcExpression, params, end)
    cfg
  }

  /**
   * Making a control flow graph for a sequence of statements: s1 ; s2 ; ... ; sn
   * makeCFG populates variables trans, vars, and asserts using side effects.
   * The decVars parameters is there to check that all variables used have been declared. It is not related to the "vars" state variable.
   */  

  def makeCFG(e: Expression, decVars: Set[String], to: Vertex) : Vertex = e match {
    case Block(es) =>
      makeCFG(es, decVars, to)
    case _ => makeCFG(List(e), decVars, to)
  }

  def makeCFG(es: List[ASTree], decVars: Set[String], to: Vertex) : Vertex = es match {
    case Nil => to 
    case e :: rest => {
      val newV = cfg.newVertex
      var newVars = decVars
      e match {
        case VarDeclaration(name, t, value) => {
          if (cfg.variables.contains(name)) throw new RuntimeException("Variable " + name + " is declared twice")
          newVars = decVars + name
        }
        case _ =>
      }
      val restV = makeCFG(rest, newVars, to)
      e.getAnnotation match {
        case Some(Annotation("assert", List(e))) => {
          checkDeclared(decVars, e)
          addAssert(newV, e)
        }
        case None =>
        case _ => throw new RuntimeException("Annotation not supported")
      }
      e match {
        case IfThenElse(c, ifB, elseB) => {
          checkDeclared(decVars, c)
          val ifV = makeCFG(ifB, decVars, restV)
          val elseV = makeCFG(elseB, decVars, restV)
          addBranch(newV, ifV, elseV, c)
          newV
        }
        case IfThen(c, ifB) => {
          checkDeclared(decVars, c)
          val ifV = makeCFG(ifB, decVars, restV)
          addBranch(newV, ifV, restV, c)
          newV
        }
        case WhileLoop(c, body) => {
          checkDeclared(decVars, c)
          val bodyV = makeCFG(body, decVars, newV)
          addBranch(newV, bodyV, restV, c)
          newV
        }
        case DoWhileLoop(c, body) => {
          checkDeclared(decVars, c)
          val bodyV = makeCFG(body, decVars, newV)
          addBranch(newV, bodyV, restV, c)
          bodyV
        }
        case VarDeclaration(name, t, value) => {
          checkDeclared(decVars, value)
          cfg.variables += name
          val v = Variable(name).stype(t)
          addAssignment(newV, restV, v, value)
          newV
        }
        case Assignment(lhs, rhs) => {
          checkDeclared(decVars, lhs)
          checkDeclared(decVars, rhs)
          addAssignment(newV, restV, lhs, rhs)
          newV
        }
        case FunctionCall("sc_assert", List(e)) => { //only one parameter allowed
          checkDeclared(decVars, e)
          addAssert(newV, e)
          fail(newV, restV, e)
          newV
        }
        case FunctionCall("sc_assume", List(e)) => { //only one parameter allowed
          checkDeclared(decVars, e)
          addAssume(newV, restV, e)
          newV
        }
        case FunctionCall("sc_havoc", List(v@Variable(_,_))) => { //only one variable allowed
          checkDeclared(decVars, v)
          addHavoc(newV, restV, v)
          newV
        }
        case _ => {
          throw new RuntimeException("Expression not handled in CFG generation: " + e)
        }
      }
    }
  }

  def checkDeclared(decVars:Set[String], e:Expression) = freeVars(e).foreach { v => 
    if (!decVars.contains(v)) throw new RuntimeException("Variable " + v + " has not been declared")
  } 

  def addAssert(v:Vertex, e:Expression) = 
    cfg.asserts += (v -> e)

  def fail(from:Vertex, to:Vertex, cond:Expression) = {
    cfg += (from, Assume(simplify(cond)), to)
    cfg += (from, Assume(simplify(Not(cond))), cfg.error)
  }

  def addHavoc(from:Vertex, to:Vertex, v:Variable) = {
    cfg += (from, Havoc(v), to)
  }

  def addAssume(from:Vertex, to:Vertex, e:Expression) = {
    cfg += (from, Assume(simplify(e)), to)
  }

  def addAssignment(from:Vertex, to:Vertex, lhs:Expression, rhs:Expression) = {
    cfg += (from, Assign(lhs, rhs), to)
  }

  def addBranch(from:Vertex, ifVertex:Vertex, elseVertex:Vertex, c:Expression) = {
    cfg += (from, Assume(simplify(c)), ifVertex)
    cfg += (from, Assume(simplify(Not(c))), elseVertex)
  }
}
